from enum import Enum
from typing import get_type_hints, Any, Union, List

from guniflask.data_model.typing import parse_json
from guniflask.orm.base_model import BaseModelMixin


class DataModel:
    def __init__(self, **data: Any):
        cls = self.__class__
        for k in get_type_hints(cls):
            if k in data:
                setattr(self, k, data[k])
            elif not hasattr(self, k):
                setattr(self, k, None)

    def to_dict(self) -> dict:
        cls = self.__class__
        d = {}
        for k in get_type_hints(cls):
            d[k] = self._get_value(getattr(self, k, None))
        return d

    def _get_value(self, v: Any) -> Any:
        if isinstance(v, DataModel):
            return v.to_dict()
        if isinstance(v, dict):
            return {_k: self._get_value(_v) for _k, _v in v.items()}
        if isinstance(v, (list, set, tuple)):
            return [self._get_value(_v) for _v in v]
        if isinstance(v, Enum):
            return v.value
        return v

    @classmethod
    def from_dict(cls, obj: dict) -> 'DataModel':
        return parse_json(obj, cls)

    @classmethod
    def from_orm(
            cls,
            obj: BaseModelMixin,
            ignore: Union[str, List] = None,
            only: Union[str, List] = None,
            include: Union[str, List] = None,
    ) -> 'DataModel':
        return cls.from_dict(
            obj.to_dict(
                ignore=ignore,
                only=only,
                include=include,
            )
        )



from typing import Union

from flask import request
from werkzeug.local import LocalProxy

from guniflask.config import settings
from guniflask.security import JwtManager, SecurityContext
from guniflask.security.authentication import Authentication
from guniflask.security.preauth_token import PreAuthenticatedToken
from guniflask.security_config import SecurityConfigurer, HttpSecurityBuilder
from guniflask.web import RequestFilter

jwt_manager: Union[JwtManager, LocalProxy] = LocalProxy(lambda: settings['jwt_manager'])


class BearerTokenExtractor:
    def extract(self) -> Authentication:
        token = self._extract_token_from_header()
        if token is None:
            token = self._extract_token_from_query()
        if token is not None:
            return PreAuthenticatedToken(token)

    @staticmethod
    def _extract_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.lower().startswith('bearer'):
            return auth.split(' ', 1)[1]

    @staticmethod
    def _extract_token_from_query():
        token = request.args.get('access_token')
        if token is not None:
            return token


class JwtConfigurer(SecurityConfigurer):
    def __init__(self, jwt=None):
        super().__init__()
        self.jwt_filter = None
        if isinstance(jwt, dict):
            settings['jwt_manager'] = JwtManager(**jwt)
            self.jwt_filter = JwtFilter()

    def configure(self, http: HttpSecurityBuilder):
        if self.jwt_filter:
            http.add_request_filter(self.jwt_filter)


class JwtFilter(RequestFilter):
    def __init__(self):
        self.token_extractor = BearerTokenExtractor()

    def before_request(self):
        auth = self.token_extractor.extract()
        if auth is not None:
            user_auth = jwt_manager.authenticate(auth)
            SecurityContext.set_authentication(user_auth)



from typing import List, Union
from urllib.parse import urlencode, urlsplit, urlunsplit

import dns.message
import dns.query
import dns.rdatatype
import requests

from guniflask.service_discovery.discovery_client import DiscoveryClient
from guniflask.service_discovery.errors import ServiceDiscoveryError
from guniflask.service_discovery.load_balancer_client import LoadBalancerClient
from guniflask.service_discovery.service_instance import ServiceInstance


class ConsulClientError(ServiceDiscoveryError):
    pass


class ConsulClient(DiscoveryClient, LoadBalancerClient):
    api_version = 'v1'

    def __init__(self, host: str = '127.0.0.1', port: int = 8500, dns_port: int = 8600, scheme: str = 'http'):
        self.host = host
        self.port = port
        self.dns_port = dns_port
        self.scheme = scheme
        self.session = requests.Session()
        self.base_url = f'{scheme}://{host}:{port}/{self.api_version}'

    def register_service(self, name: str,
                         service_id: str = None,
                         tags: List[str] = None,
                         address: str = None,
                         port: int = None,
                         check: dict = None):
        api_path = '/agent/service/register'
        args = {}
        if check is not None:
            args['replace-existing-checks'] = 'true'
        if len(args) > 0:
            api_path = f'{api_path}?{urlencode(args)}'

        data = {
            'Name': name,
            'ID': service_id,
            'Tags': tags,
            'Address': address,
            'Port': port,
            'Check': check
        }
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.put(url, json=data)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)

    def deregister_service(self, service_id: str):
        api_path = f'/agent/service/deregister/{service_id}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.put(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)

    def get_service_by_id(self, service_id: str):
        api_path = f'/agent/service/{service_id}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.get(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)
        return resp.json()

    @staticmethod
    def http_check(name: str, url: str, check_id: str = None, interval: str = None, deregister_after: str = None):
        return {'Name': name,
                'CheckID': check_id,
                'HTTP': url,
                'Interval': interval,
                'DeregisterCriticalServiceAfter': deregister_after}

    def get_service_instances(self, service_name: str) -> List[ServiceInstance]:
        api_path = f'/agent/health/service/name/{service_name}'
        url = f'{self.base_url}{api_path}'
        try:
            resp = self.session.get(url)
            resp.raise_for_status()
        except Exception as e:
            raise ConsulClientError(e)
        data = resp.json()
        services = []
        for d in data:
            if d['AggregatedStatus'] == 'passing':
                s = d['Service']
                services.append(ServiceInstance(service_id=s['ID'],
                                                host=s['Address'],
                                                port=s['Port']))
        return services

    def choose(self, service_name: str) -> Union[ServiceInstance, None]:
        request = dns.message.make_query(f'{service_name}.service.consul', dns.rdatatype.SRV)
        response = dns.query.udp(request, self.host, port=self.dns_port)
        if len(response.answer) > 0:
            answer = response.answer[0]
            port = None
            target = None
            for k in answer:
                port = k.port
                target = k.target
                break
            if target is not None:
                for additional in response.additional:
                    if additional.name == target:
                        for k in additional:
                            return ServiceInstance(host=k.address, port=port)

    def reconstruct_url(self, service_instance: ServiceInstance, original_url: str) -> str:
        result = urlsplit(original_url)
        result = result._replace(netloc=f'{service_instance.host}:{service_instance.port}')
        return urlunsplit(result)



import inspect
from abc import ABCMeta, abstractmethod
from typing import Type

from guniflask.security.authentication import Authentication
from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.user_details import UserDetails


class UserDetailsAuthenticationProvider(AuthenticationProvider, metaclass=ABCMeta):

    @abstractmethod
    def retrieve_user(self, username: str, authentication: UserAuthentication) -> UserDetails:
        pass  # pragma: no cover

    @abstractmethod
    def check_authentication(self, user_details: UserDetails, authentication: UserAuthentication):
        pass  # pragma: no cover

    def authenticate(self, authentication: Authentication):
        assert isinstance(authentication, UserAuthentication)
        username = authentication.name
        user = self.retrieve_user(username, authentication)
        self.check_authentication(user, authentication)
        return UserAuthentication(user, credentials=authentication.credentials, authorities=user.authorities)

    def supports(self, authentication_cls: Type[Authentication]) -> bool:
        return inspect.isclass(authentication_cls) and issubclass(authentication_cls, UserAuthentication)



class AuthenticationError(Exception):
    """
    Authentication exception
    """


class BadCredentialsError(AuthenticationError):
    """
    Bad credentials exception
    """


class ProviderNotFoundError(AuthenticationError):
    """
    Provider not found exception
    """


class UsernameNotFoundError(AuthenticationError):
    """
    Username not found exception
    """


class InsufficientAuthenticationError(AuthenticationError):
    """
    Insufficient authentication exception
    """


class InvalidTokenError(AuthenticationError):
    """
    Invalid token exception
    """



import re

_camelcase_invalid_chars = re.compile(r'[^a-zA-Z\d]')

_camelcase_split_chars = re.compile(r'(.)([A-Z][a-z]+)')
_camelcase_split_chars2 = re.compile(r'([a-z0-9])([A-Z])')


def string_camelcase(s):
    a = _camelcase_invalid_chars.split(s)
    return ''.join([(i[0].upper() + i[1:]) for i in a if i])


def string_lowercase_hyphen(s):
    return string_lowercase_underscore(s).replace('_', '-')


def string_lowercase_underscore(s):
    s = string_camelcase(s)
    s = _camelcase_split_chars.sub(r'\1_\2', s)
    return _camelcase_split_chars2.sub(r'\1_\2', s).lower()


def string_uppercase_underscore(s):
    return string_lowercase_underscore(s).upper()


def comma_delimited_list_to_set(s):
    return set(comma_delimited_list_to_array(s))


def comma_delimited_list_to_array(s):
    return s.split(',')



import inspect

from guniflask.annotation import Annotation, AnnotationUtils


class AsyncRun(Annotation):
    def __init__(self, executor: str = None):
        super().__init__(executor=executor)


def async_run(executor: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, AsyncRun(executor=executor))
        return func

    if inspect.isclass(executor) or inspect.isfunction(executor):
        f = executor
        executor = None
        return wrap_func(f)
    return wrap_func


class Scheduled(Annotation):
    def __init__(self, cron: str = None, interval: int = None, initial_delay: int = None):
        super().__init__(cron=cron, interval=interval, initial_delay=initial_delay)


def scheduled(cron: str = None, interval: int = None, initial_delay: int = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Scheduled(cron=cron,
                                                       interval=interval,
                                                       initial_delay=initial_delay))
        return func

    if inspect.isclass(cron) or inspect.isfunction(cron):
        f = cron
        cron = None
        return wrap_func(f)
    return wrap_func



from guniflask.security_config.security_builder import SecurityBuilder


class SecurityConfigurer:
    def __init__(self):
        self._security_builder: SecurityBuilder = None

    def init(self, builder: SecurityBuilder):
        pass

    def configure(self, builder: SecurityBuilder):
        pass

    @property
    def builder(self):
        return self._security_builder

    @builder.setter
    def builder(self, value: SecurityBuilder):
        self._security_builder = value



from .bind_annotation import blueprint
from .bind_annotation import delete_route
from .bind_annotation import get_route
from .bind_annotation import patch_route
from .bind_annotation import post_route
from .bind_annotation import put_route
from .bind_annotation import route
from .context import WebApplicationContext
from .cors import CorsFilter
from .filter_annotation import after_request
from .filter_annotation import app_after_request
from .filter_annotation import app_before_request
from .filter_annotation import app_error_handler
from .filter_annotation import before_request
from .filter_annotation import error_handler
from .request_annotation import CookieValue
from .request_annotation import FilePart
from .request_annotation import FormValue
from .request_annotation import PathVariable
from .request_annotation import RequestBody
from .request_annotation import RequestHeader
from .request_annotation import RequestParam
from .request_filter import RequestFilter
from .request_filter import RequestFilterChain



class BeansError(Exception):
    pass


class NoSuchBeanDefinitionError(BeansError):
    def __init__(self, bean_name):
        self.bean_name = bean_name
        super().__init__(f'No bean named "{bean_name}" available')


class NoUniqueBeanDefinitionError(BeansError):
    def __init__(self, bean_type):
        self.bean_type = bean_type
        super().__init__(f'"{bean_type}" is not a unique bean type')


class BeanDefinitionStoreError(BeansError):
    pass


class BeanCreationError(BeansError):
    def __init__(self, bean_name, message=None):
        self.bean_name = bean_name
        super().__init__(message or f'Error creating bean with name "{bean_name}"')


class BeanCurrentlyInCreationError(BeanCreationError):
    def __init__(self, bean_name):
        super().__init__(bean_name, message=f'Bean named "{bean_name}" is currently in creation: '
                                            'Is there an unresolvable circular reference?"')


class BeanNotOfRequiredTypeError(BeansError):
    def __init__(self, bean_name, required_type, actual_type):
        self.bean_name = bean_name
        self.required_type = required_type
        self.actual_type = actual_type
        super().__init__(f'Bean named "{bean_name}" is expected to be of type "{required_type}"'
                         f' but was actually of type {actual_type}')


class BeanTypeNotDeclaredError(BeansError):
    def __init__(self, bean_name):
        self.bean_name = bean_name
        super().__init__(f'The type of bean named "{bean_name}" is not declared')


class BeanTypeNotAllowedError(BeansError):
    def __init__(self, bean_name, bean_type):
        self.bean_name = bean_name
        self.bean_type = bean_type
        super().__init__(f'The type "{bean_type}" of bean named "{bean_name}" is not a allowed')



from abc import ABCMeta, abstractmethod
from urllib.parse import urlsplit

from requests import Session, Request

from guniflask.service_discovery.service_instance import ServiceInstance


class LoadBalancerClient(metaclass=ABCMeta):

    @abstractmethod
    def choose(self, service_name: str) -> ServiceInstance:
        pass  # pragma: no cover

    @abstractmethod
    def reconstruct_url(self, service_instance: ServiceInstance, original_url: str) -> str:
        pass  # pragma: no cover


class RestClient(Session):

    def __init__(self, load_balancer_client: LoadBalancerClient):
        assert load_balancer_client is not None, 'LoadBalancerClient is required'
        self.load_balancer_client = load_balancer_client
        super().__init__()

    def prepare_request(self, request: Request):
        result = urlsplit(request.url)
        if '.' not in result.netloc:
            service_name = result.netloc
            service_instance = self.load_balancer_client.choose(service_name)
            if service_instance is not None:
                request.url = self.load_balancer_client.reconstruct_url(service_instance, request.url)
        return super().prepare_request(request)



import logging
from abc import ABCMeta, abstractmethod

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger

from guniflask.config.app_settings import settings, Settings
from guniflask.context.annotation import configuration, bean, condition_on_setting, include
from guniflask.distributed.local_lock import ServiceLock
from guniflask.service_discovery.consul import ConsulClient, ConsulClientError
from guniflask.service_discovery.discovery_client import DiscoveryClient
from guniflask.service_discovery.health_endpoint import HealthEndpoint
from guniflask.service_discovery.load_balancer_client import LoadBalancerClient, RestClient

log = logging.getLogger(__name__)


class ServiceDiscoveryConfigurer(metaclass=ABCMeta):
    @abstractmethod
    def configure(self, service_name: str, app_settings: Settings) -> bool:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def discovery_client(self) -> DiscoveryClient:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def load_balancer_client(self) -> LoadBalancerClient:
        pass  # pragma: no cover


@condition_on_setting('guniflask.consul')
@configuration
class ConsulConfigurer(ServiceDiscoveryConfigurer):

    def __init__(self):
        self._discovery_client: DiscoveryClient = None
        self._load_balancer_client: LoadBalancerClient = None
        self._register_scheduler = None
        self._service_name: str = None

    @property
    def discovery_client(self) -> DiscoveryClient:
        return self._discovery_client

    @property
    def load_balancer_client(self) -> LoadBalancerClient:
        return self._load_balancer_client

    def configure(self, service_name: str, app_settings: Settings):
        config = app_settings.get_by_prefix('guniflask.consul')
        self._service_name = service_name
        self._register_scheduler = BackgroundScheduler()
        self._register_scheduler.start(paused=False)
        self._consul_register(config, app_settings)

    def _consul_register(self, config: dict, app_settings: Settings):
        consul = self._consul_client(config)
        self._discovery_client = consul
        self._load_balancer_client = consul

        job_id = 'consul_register'
        if self._register_scheduler.get_job(job_id) is None:
            self._do_consul_register(consul, app_settings)
            trigger = IntervalTrigger(minutes=1)
            self._register_scheduler.add_job(self._do_consul_register,
                                             args=(consul, app_settings),
                                             id=job_id,
                                             trigger=trigger)

    def _consul_client(self, config: dict) -> ConsulClient:
        client_config = {}
        for i in ['host', 'port', 'dns_port', 'scheme']:
            if i in config:
                client_config[i] = config[i]
        return ConsulClient(**client_config)

    def _do_consul_register(self, consul: ConsulClient, app_settings: Settings):
        local_ip = app_settings['ip_address']
        port = app_settings['port']
        service_id = f'{app_settings["app_name"]}-{local_ip}-{port}'
        heath_url = f'http://{local_ip}:{port}/health?' \
                    f'app_id={app_settings["app_id"]}'
        try:
            consul.register_service(app_settings['app_name'],
                                    service_id=service_id,
                                    address=local_ip,
                                    port=port,
                                    check=consul.http_check(service_id,
                                                            heath_url,
                                                            check_id=service_id,
                                                            interval='10s',
                                                            deregister_after='10m'))
        except ConsulClientError as e:
            log.error('Failed to register service to Consul at %s:%s: %s', consul.host, consul.port, e)


@configuration
@include(ConsulConfigurer)
class ServiceDiscoveryConfiguration:

    def __init__(self, service_discovery_configurer: ServiceDiscoveryConfigurer = None):
        self._service_discovery_configurer = service_discovery_configurer

        self._register_lock = ServiceLock('__service_registration_lock')
        self._auto_register()

    @bean
    def heath_endpoint(self) -> HealthEndpoint:
        endpoint = HealthEndpoint()
        return endpoint

    @bean
    def discovery_client(self) -> DiscoveryClient:
        if self._service_discovery_configurer:
            return self._service_discovery_configurer.discovery_client

    @bean
    def load_balancer_client(self) -> LoadBalancerClient:
        if self._service_discovery_configurer:
            return self._service_discovery_configurer.load_balancer_client

    @bean
    def load_balanced_request(self) -> RestClient:
        if self._service_discovery_configurer:
            load_balancer_client = self._service_discovery_configurer.load_balancer_client
            if load_balancer_client is not None:
                return RestClient(load_balancer_client)

    def _auto_register(self):
        if not self._register_lock.acquire():
            return

        app_settings = settings._get_current_object()
        service_name = app_settings['app_name']
        if self._service_discovery_configurer is not None:
            self._service_discovery_configurer.configure(service_name, app_settings)



from abc import abstractmethod, ABCMeta

from guniflask.beans.definition import BeanDefinition
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.utils.string import string_lowercase_underscore


class BeanNameGenerator(metaclass=ABCMeta):
    @abstractmethod
    def generate_bean_name(self, bean_definition: BeanDefinition,
                           registry: BeanDefinitionRegistry) -> str:
        pass  # pragma: no cover


class DefaultBeanNameGenerator(BeanNameGenerator):

    def generate_bean_name(self, bean_definition: BeanDefinition,
                           bean_definition_registry: BeanDefinitionRegistry) -> str:
        source = bean_definition.source
        name = source.__name__
        bean_name = string_lowercase_underscore(name)
        return bean_name



import inspect
from abc import ABCMeta, abstractmethod

from guniflask.annotation import AnnotationMetadata, AnnotationUtils
from guniflask.beans.definition import BeanDefinition
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.context.annotation import Component
from guniflask.context.annotation_config_utils import AnnotationConfigUtils
from guniflask.context.bean_name_generator import AnnotationBeanNameGenerator
from guniflask.context.condition_evaluator import ConditionEvaluator
from guniflask.utils.path import walk_modules


class AnnotationConfigRegistry(metaclass=ABCMeta):
    @abstractmethod
    def register(self, *annotated_elements):
        pass

    @abstractmethod
    def scan(self, *base_modules):
        pass


class AnnotatedBeanDefinitionReader:
    def __init__(self, registry: BeanDefinitionRegistry):
        self._registry = registry
        self._bean_name_generator = AnnotationBeanNameGenerator()
        self._condition_evaluator = ConditionEvaluator(registry)
        AnnotationConfigUtils.register_annotation_config_processors(self._registry)

    def register(self, *annotated_elements):
        for e in annotated_elements:
            self.register_bean(e)

    def register_bean(self, annotated_element, name=None):
        if self._condition_evaluator.should_skip(AnnotationUtils.get_annotation_metadata(annotated_element)):
            return
        bean_definition = BeanDefinition(annotated_element)
        bean_name = name or self._bean_name_generator.generate_bean_name(bean_definition, self._registry)
        self._registry.register_bean_definition(bean_name, bean_definition)

    def set_bean_name_generator(self, bean_name_generator):
        self._bean_name_generator = bean_name_generator


class ModuleBeanDefinitionScanner:
    def __init__(self, registry: BeanDefinitionRegistry):
        self._registry = registry
        self.include_annotation_config = True
        self._bean_name_generator = AnnotationBeanNameGenerator()
        self._condition_evaluator = ConditionEvaluator(registry)

    def scan(self, *base_modules):
        self._scan(*base_modules)
        if self.include_annotation_config:
            AnnotationConfigUtils.register_annotation_config_processors(self._registry)

    def set_bean_name_generator(self, bean_name_generator):
        self._bean_name_generator = bean_name_generator

    def _scan(self, *base_modules):
        for base_module in base_modules:
            for module in walk_modules(base_module):
                candidates = self._find_candidate_components(module)
                for bean_definition in candidates:
                    bean_name = self._bean_name_generator.generate_bean_name(bean_definition, self._registry)
                    self._registry.register_bean_definition(bean_name, bean_definition)

    def _find_candidate_components(self, module):
        candidates = []
        selected_id = set()
        for obj in vars(module).values():
            if inspect.isclass(obj) or inspect.isfunction(obj):
                if obj.__module__ == module.__name__:
                    annotation_metadata = AnnotationUtils.get_annotation_metadata(obj)
                    if annotation_metadata is not None and self._is_candidate_component(annotation_metadata):
                        obj_id = id(obj)
                        if obj_id not in selected_id:
                            selected_id.add(obj_id)
                            bean_definition = BeanDefinition(obj)
                            candidates.append(bean_definition)
        return candidates

    def _is_candidate_component(self, metadata: AnnotationMetadata):
        return metadata.is_annotated(Component) and not self._condition_evaluator.should_skip(metadata)



import inspect
from typing import Type

from guniflask.security.authentication import Authentication
from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.security.authentication_user_details_service import AuthenticationUserDetailsService
from guniflask.security.preauth_token import PreAuthenticatedToken


class PreAuthenticatedProvider(AuthenticationProvider):
    def __init__(self, user_details_service: AuthenticationUserDetailsService):
        self.pre_authenticated_user_details_service = user_details_service

    def authenticate(self, authentication: Authentication):
        if not self.supports(type(authentication)):
            return
        if authentication.principal is None:
            return
        if authentication.credentials is None:
            return
        user_details = self.pre_authenticated_user_details_service.load_user_details(authentication)
        result = PreAuthenticatedToken(user_details, authentication.credentials, user_details.authorities)
        result.details = authentication.details
        return result

    def supports(self, authentication_cls: Type[Authentication]) -> bool:
        return inspect.isclass(authentication_cls) and issubclass(authentication_cls, PreAuthenticatedToken)



import inspect
from collections import defaultdict

from guniflask.annotation import AnnotationUtils
from guniflask.beans.constructor_resolver import ConstructorResolver
from guniflask.beans.factory import BeanFactory
from guniflask.beans.factory import BeanFactoryAware
from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.context.annotation import Autowired


class AutowiredAnnotationBeanPostProcessor(BeanPostProcessor, BeanFactoryAware):

    def __init__(self):
        self._autowired_methods = defaultdict(list)
        self._bean_factory: BeanFactory = None
        self._constructor_resolver: ConstructorResolver = None

    def set_bean_factory(self, bean_factory: BeanFactory):
        self._bean_factory = bean_factory
        self._constructor_resolver = ConstructorResolver(bean_factory)

    def post_process_before_instantiation(self, bean_type: type, bean_name: str):
        for m in dir(bean_type):
            method = getattr(bean_type, m)
            if inspect.ismethod(method) or inspect.isfunction(method):
                a = AnnotationUtils.get_annotation(method, Autowired)
                if a is not None:
                    self._autowired_methods[bean_name].append(m)

    def post_process_before_initialization(self, bean, bean_name: str):
        if bean_name in self._autowired_methods:
            for m in self._autowired_methods[bean_name]:
                if hasattr(bean, m):
                    method = getattr(bean, m)
                    self._constructor_resolver.instantiate(method)



from guniflask.security.authentication import Authentication
from guniflask.security.authentication_user_details_service import AuthenticationUserDetailsService
from guniflask.security.user_details import UserDetails
from guniflask.security.user_details_service import UserDetailsService


class UserDetailsByNameService(AuthenticationUserDetailsService):
    def __init__(self, user_details_service: UserDetailsService):
        self._user_details_service = user_details_service

    def load_user_details(self, token: Authentication) -> UserDetails:
        return self._user_details_service.load_user_by_username(token.name)



from abc import ABCMeta, abstractmethod
from typing import List

from guniflask.service_discovery.service_instance import ServiceInstance


class DiscoveryClient(metaclass=ABCMeta):

    @abstractmethod
    def get_service_instances(self, service_name: str) -> List[ServiceInstance]:
        pass  # pragma: no cover



import datetime as dt


def convert_to_datetime(s):
    if isinstance(s, int) or (isinstance(s, str) and s.isdigit()):
        return dt.datetime.fromtimestamp(int(s), tz=dt.timezone.utc).astimezone()
    if isinstance(s, str):
        if 'GMT' in s:
            return dt.datetime.strptime(s, '%a, %d %b %Y %H:%M:%S GMT').replace(tzinfo=dt.timezone.utc).astimezone()
    return s



import io
import json
import os
import time
from os.path import dirname, abspath, join

import pytest
from flask.testing import FlaskClient

from guniflask.test import app_test_client


@pytest.fixture
def rest_client():
    os.environ['GUNIFLASK_HOME'] = join(dirname(abspath(__file__)), 'rest_app')
    with app_test_client() as client:
        client.application.config['TESTING'] = True
        yield client
        clear_env()


@pytest.fixture
def scheduler_client():
    os.environ['GUNIFLASK_HOME'] = join(dirname(abspath(__file__)), 'scheduler_app')
    with app_test_client() as client:
        client.application.config['TESTING'] = True
        yield client
        clear_env()


def clear_env():
    for k in os.environ:
        if k.startswith('GUNIFLASK_'):
            os.environ.pop(k)


def get_json(resp):
    assert resp.status_code == 200
    return resp.get_json()


class TestRestApp:
    def test_health_info(self, rest_client: FlaskClient):
        data = get_json(rest_client.get('/health'))
        assert data == {'status': 'UP'}

    def test_get_default_setting(self, rest_client: FlaskClient):
        data = get_json(rest_client.get('/settings/app_name'))
        assert data['app_name'] == 'rest_app'

        data = get_json(rest_client.get('/settings/debug'))
        assert data['debug'] is True

    def test_math_service(self, rest_client: FlaskClient):
        data = get_json(rest_client.get('/api/math/add?x=1&y=2'))
        assert data['result'] == 3

    def test_query_param(self, rest_client: FlaskClient):
        data = get_json(rest_client.get('/get/int?x=1'))
        assert data['x'] == 1

        data = get_json(rest_client.get('/get/str?x=1'))
        assert data['x'] == '1'

        data = get_json(rest_client.get('/get/default'))
        assert data['x'] == 'default'

    def test_post_dict(self, rest_client: FlaskClient):
        data = get_json(
            rest_client.post(
                '/post/dict',
                json=dict(
                    a=1,
                    b='1',
                ),
            )
        )
        assert data['a'] == 1
        assert data['b'] == '1'

    def test_post_list(self, rest_client: FlaskClient):
        data = get_json(
            rest_client.post(
                '/post/list',
                data=json.dumps([1, '1']),
                content_type='application/json',
            )
        )
        assert data == [1, '1']

    def test_post_form(self, rest_client: FlaskClient):
        data = get_json(
            rest_client.post(
                '/post/form-int',
                data='x=1',
                content_type='application/x-www-form-urlencoded',
            )
        )
        assert data == {'x': 1}

        data = get_json(
            rest_client.post(
                '/post/form-str',
                data='x=1',
                content_type='application/x-www-form-urlencoded',
            )
        )
        assert data == {'x': '1'}

    def test_post_file(self, rest_client: FlaskClient):
        data = get_json(
            rest_client.post(
                '/post/file-bytes',
                data={
                    'x': (io.BytesIO(b'file'), 'test.txt')
                },
                content_type='multipart/form-data',
            )
        )
        assert data['x'] == 'file'

    def test_post_cookie(self, rest_client: FlaskClient):
        rest_client.set_cookie('localhost', 'x', '1')
        data = get_json(
            rest_client.post(
                '/post/cookie-int',
            )
        )
        assert data['x'] == 1

        data = get_json(
            rest_client.post(
                '/post/cookie-str',
            )
        )
        assert data['x'] == '1'

    def test_post_header(self, rest_client: FlaskClient):
        rest_client.set_cookie('localhost', 'x', '1')
        data = get_json(
            rest_client.post(
                '/post/header-int',
                headers={'x': '1'},
            )
        )
        assert data['x'] == 1

        data = get_json(
            rest_client.post(
                '/post/header-str',
                headers={'x': '1'},
            )
        )
        assert data['x'] == '1'

    def test_login(self, rest_client: FlaskClient):
        data = get_json(
            rest_client.post(
                '/login?username=root&password=123456',
            )
        )
        assert data['username'] == 'root'
        assert 'access_token' in data
        token = data['access_token']
        data = get_json(
            rest_client.get(
                '/accounts/root',
                headers={'Authorization': 'Bearer ' + token},
            )
        )
        assert data['username'] == 'root'
        assert data['authorities'] == ['role_admin']


class TestSchedulerApp:
    def test_schedule_task(self, scheduler_client: FlaskClient):
        scheduler_client.post('/async-add?x=1&y=2')
        time.sleep(1.2)
        data = get_json(scheduler_client.get('/scheduled'))
        assert data['result'] is True
        data = get_json(scheduler_client.get('/async-add'))
        assert data['result'] == 3



from typing import Any


class BeanPostProcessor:
    def post_process_before_instantiation(self, bean_type: type, bean_name: str):
        return None

    def post_process_before_initialization(self, bean: Any, bean_name: str) -> Any:
        return bean

    def post_process_after_initialization(self, bean: Any, bean_name: str) -> Any:
        return bean



from guniflask.annotation import AnnotationUtils
from guniflask.beans.definition import BeanDefinition
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.beans.name_generator import DefaultBeanNameGenerator, BeanNameGenerator
from guniflask.context.annotation import Component


class AnnotationBeanNameGenerator(BeanNameGenerator):
    def __init__(self):
        self._default_generator = DefaultBeanNameGenerator()

    def generate_bean_name(self, bean_definition: BeanDefinition,
                           registry: BeanDefinitionRegistry) -> str:
        bean_name = None
        annotations = AnnotationUtils.get_annotations(bean_definition.source)
        for annotation in annotations:
            if isinstance(annotation, Component):
                name = annotation['name']
                if name is not None:
                    if bean_name is not None and name != bean_name:
                        raise ValueError('Annotations suggest inconsistent component names: '
                                         f'"{bean_name}" versus {name}')
                    bean_name = name
        if bean_name is not None:
            return bean_name
        return self._default_generator.generate_bean_name(bean_definition, registry)



# Database URI when using Flask-SQLAlchemy, example: mysql+pymysql://username:password@server/db?charset=utf8mb4
# SQLALCHEMY_DATABASE_URI = ''


# guniflask configuration
guniflask = dict(
    cors=True,
)



import inspect

from guniflask.annotation import Annotation, AnnotationUtils
from guniflask.context.annotation import Component


class Blueprint(Component):
    def __init__(self, url_prefix: str = None, **options):
        super().__init__()
        self['url_prefix'] = url_prefix
        self['options'] = options


def blueprint(url_prefix: str = None, **options):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Blueprint(url_prefix=url_prefix, **options))
        return func

    if inspect.isclass(url_prefix) or inspect.isfunction(url_prefix):
        f = url_prefix
        url_prefix = None
        return wrap_func(f)
    return wrap_func


class Route(Annotation):
    def __init__(self, rule: str = None, **options):
        super().__init__(rule=rule, options=options)


def route(rule: str = None, **options):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Route(rule=rule, **options))
        return func

    if inspect.isfunction(rule) or inspect.isclass(rule):
        f = rule
        rule = None
        return wrap_func(f)
    return wrap_func


def get_route(rule: str = None, **options):
    options['methods'] = ['GET']
    return route(rule=rule, **options)


def post_route(rule: str = None, **options):
    options['methods'] = ['POST']
    return route(rule=rule, **options)


def put_route(rule: str = None, **options):
    options['methods'] = ['PUT']
    return route(rule=rule, **options)


def patch_route(rule: str = None, **options):
    options['methods'] = ['PATCH']
    return route(rule=rule, **options)


def delete_route(rule: str = None, **options):
    options['methods'] = ['DELETE']
    return route(rule=rule, **options)



import inspect

from guniflask.beans.errors import NoUniqueBeanDefinitionError, BeansError
from guniflask.beans.factory import BeanFactory
from guniflask.data_model.typing import inspect_args, analyze_arg_type


class ConstructorResolver:
    def __init__(self, bean_factory: BeanFactory):
        self._bean_factory = bean_factory

    def instantiate(self, func):
        args, hints = inspect_args(func)

        args_value = []
        kwargs_value = {}
        for name, default in args.items():
            required_type = hints.get(name)
            v = self._resolve_arg(name, required_type)
            if default is inspect._empty:
                if v is None:
                    raise ValueError(f'The argument named "{name}" cannot be resolved')
                args_value.append(v)
            else:
                if v is None:
                    kwargs_value[name] = default
                else:
                    kwargs_value[name] = v

        return func(*args_value, **kwargs_value)

    def _resolve_arg(self, arg, required_type):
        bean = None
        arg_ = analyze_arg_type(required_type)

        if arg_.outer_type is None:
            try:
                bean = self._bean_factory.get_bean(arg)
            except BeansError:
                pass
        elif inspect.isclass(arg_.outer_type):
            candidates = self._bean_factory.get_beans_of_type(arg_.outer_type)
            if arg_.is_list():
                bean = list(candidates.values())
            elif arg_.is_set():
                bean = set(candidates.values())
            elif arg_.is_singleton():
                if len(candidates) == 1:
                    bean = list(candidates.values())[0]
                elif len(candidates) > 1:
                    if arg in candidates:
                        bean = candidates[arg]
                    else:
                        raise NoUniqueBeanDefinitionError(required_type)
        return bean



import os
import re
import sys
from os.path import join, isfile, isdir

from dotenv import load_dotenv


def set_app_default_env():
    home_dir = os.environ.get('GUNIFLASK_HOME')
    if not home_dir:
        home_dir = os.getcwd()
        os.environ['GUNIFLASK_HOME'] = home_dir
    if home_dir not in sys.path:
        sys.path.append(home_dir)
    if not os.environ.get('GUNIFLASK_CONF_DIR'):
        os.environ['GUNIFLASK_CONF_DIR'] = join(home_dir, 'conf')


def load_env(fname):
    if fname is None or not isfile(fname):
        raise FileNotFoundError(f"Cannot find env file '{fname}'")
    load_dotenv(fname)


def load_profile_env(conf_dir, profiles: str = None):
    base_file = join(conf_dir, 'app.env')
    if isfile(base_file):
        load_env(base_file)
    if profiles:
        profiles = profiles.split(',')
        for profile in reversed(profiles):
            if profile:
                p_file = join(conf_dir, f'app_{profile}.env')
                if isfile(p_file):
                    load_env(p_file)


def load_app_env():
    set_app_default_env()
    conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
    active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES')
    load_profile_env(conf_dir, active_profiles)


app_name_regex = re.compile(r'[a-zA-Z0-9_]+')


def infer_app_name(home_dir):
    candidates = []
    for d in os.listdir(home_dir):
        if app_name_regex.fullmatch(d) \
                and _find_py_module(join(home_dir, d), '__init__') \
                and _find_py_module(join(home_dir, d), 'app'):
            candidates.append(d)
    if len(candidates) == 0:
        return None
    if len(candidates) > 1:
        raise RuntimeError(f'Cannot infer the app name, candidates: {candidates}')
    return candidates[0]


def _find_py_module(path, name):
    if not isdir(path):
        return False
    for s in os.listdir(path):
        if s.startswith(name) and (s.endswith('.py') or s.endswith('.so')):
            return True
    return False


def app_name_from_env():
    if not os.environ.get('GUNIFLASK_APP_NAME'):
        app_name = infer_app_name(os.environ.get('GUNIFLASK_HOME'))
        if app_name:
            os.environ['GUNIFLASK_APP_NAME'] = app_name
    return os.environ.get('GUNIFLASK_APP_NAME')



from guniflask.security.dao_authentication_provider import DaoAuthenticationProvider
from guniflask.security.password_encoder import PasswordEncoder
from guniflask.security.user_details_service import UserDetailsService
from guniflask.security_config.provider_manager_builder import ProviderManagerBuilder
from guniflask.security_config.user_details_aware_configurer import UserDetailsAwareConfigurer


class DaoAuthenticationConfigurer(UserDetailsAwareConfigurer):

    def __init__(self, user_details_service: UserDetailsService):
        super().__init__()
        self._user_details_service = user_details_service
        self._provider = DaoAuthenticationProvider()
        self._provider.user_details_service = user_details_service

    @property
    def user_details_service(self) -> UserDetailsService:
        return self._user_details_service

    def configure(self, builder: ProviderManagerBuilder):
        builder.with_authentication_provider(self._provider)

    def with_password_encoder(self, password_encoder: PasswordEncoder) -> 'DaoAuthenticationConfigurer':
        self._provider.password_encoder = password_encoder
        return self



from typing import Any, List

from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.security.provider_manager import ProviderManager
from guniflask.security.user_details_service import UserDetailsService
from guniflask.security_config.configured_security_builder import ConfiguredSecurityBuilder
from guniflask.security_config.dao_authentication_configurer import DaoAuthenticationConfigurer
from guniflask.security_config.provider_manager_builder import ProviderManagerBuilder


class AuthenticationManagerBuilder(ConfiguredSecurityBuilder, ProviderManagerBuilder):
    def __init__(self):
        super().__init__()
        self._parent_authentication_manager: AuthenticationManager = None
        self._authentication_providers: List[AuthenticationProvider] = []
        self._default_user_details_service: UserDetailsService = None

    def _perform_build(self) -> Any:
        provider_manager = ProviderManager(parent=self._parent_authentication_manager,
                                           providers=self._authentication_providers)
        return provider_manager

    @property
    def default_user_details_service(self) -> UserDetailsService:
        return self._default_user_details_service

    def with_authentication_provider(self, provider: AuthenticationProvider) -> 'AuthenticationManagerBuilder':
        self._authentication_providers.append(provider)
        return self

    def with_parent_authentication_manager(
            self, authentication_manager: AuthenticationManager
    ) -> 'AuthenticationManagerBuilder':
        self._parent_authentication_manager = authentication_manager
        return self

    def with_user_details_service(self, user_details_service: UserDetailsService) -> DaoAuthenticationConfigurer:
        self._default_user_details_service = user_details_service
        return self.apply(DaoAuthenticationConfigurer(user_details_service))



import os
import uuid
from os.path import isfile, join

from guniflask.utils.network import get_local_ip_address


def load_config(fname, **kwargs) -> dict:
    if fname is None or not isfile(fname):
        raise FileNotFoundError(f"Cannot find configuration file '{fname}'")
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    cfg.update(kwargs)
    exec(code, cfg, cfg)
    return cfg


def load_profile_config(conf_dir, name, profiles=None, **kwargs) -> dict:
    base_file = join(conf_dir, name + '.py')
    if isfile(base_file):
        pc = load_config(base_file, **kwargs)
    else:
        pc = {}
    if profiles:
        profiles = profiles.split(',')
        for profile in reversed(profiles):
            if profile:
                pc_file = join(conf_dir, name + '_' + profile + '.py')
                if isfile(pc_file):
                    c = load_config(pc_file, **kwargs)
                    _update_config(pc, c)
        pc['active_profiles'] = list(profiles)
    return pc


def _update_config(old: dict, new: dict):
    for k, v in new.items():
        if k not in old:
            old[k] = v
        else:
            if isinstance(v, dict) and isinstance(old[k], dict):
                _update_config(old[k], v)
            else:
                old[k] = v


def load_app_settings(app_name) -> dict:
    c = {}
    conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
    active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES')
    kwargs = get_settings_from_env()
    kwargs['app_name'] = app_name
    if conf_dir:
        c = load_profile_config(conf_dir, app_name, profiles=active_profiles, **kwargs)
    if 'app_id' not in c:
        c['app_id'] = uuid.uuid4().hex
    if 'ip_address' not in c:
        c['ip_address'] = get_local_ip_address()

    s = {}
    for name in c:
        if not name.startswith('_'):
            s[name] = c[name]
    return s


def get_settings_from_env() -> dict:
    kwargs = {
        'home': os.environ.get('GUNIFLASK_HOME'),
    }
    if os.environ.get('GUNIFLASK_DEBUG'):
        kwargs['debug'] = True
    else:
        kwargs['debug'] = False
    kwargs['host'] = os.environ.get('GUNIFLASK_HOST')
    port = os.environ.get('GUNIFLASK_PORT')
    if port:
        port = int(port)
    kwargs['port'] = port
    return kwargs



from abc import ABCMeta, abstractmethod

from guniflask.security.user_details_service import UserDetailsService
from guniflask.security_config.security_configurer import SecurityConfigurer


class UserDetailsAwareConfigurer(SecurityConfigurer, metaclass=ABCMeta):
    @property
    @abstractmethod
    def user_details_service(self) -> UserDetailsService:
        pass  # pragma: no cover



from guniflask.config import Settings


def test_get():
    d = {'key': 'value'}
    c = Settings(d)
    assert len(c) == 1
    for k in c:
        assert k == 'key' and c[k] == 'value'
    assert c.get('key') == 'value'
    assert c.get('no_such_key') is None
    assert c.get('no_such_key', 'default') == 'default'
    assert c.get('key', 'default') == 'value'
    assert c['key'] == 'value'
    assert c['no_such_key'] is None
    assert ('key' in c) is True
    assert ('no_such_key' in c) is False


def test_get_bool():
    d = {'bool_true': 'true', 'bool_True': 'True', 'bool_false': 'false', 'bool_False': 'False',
         'bool_int1': '1', 'bool_int0': '0', 'bool_none': '...',
         'true': True, 'false': False}
    c = Settings(d)
    assert c.getbool('true') is True
    assert c.getbool('false') is False
    assert c.getbool('bool_true') is True
    assert c.getbool('bool_True') is True
    assert c.getbool('bool_false') is False
    assert c.getbool('bool_False') is False
    assert c.getbool('bool_int1') is True
    assert c.getbool('bool_int0') is False
    assert c.getbool('bool_none') is None
    assert c.getbool('bool_no') is None
    assert c.getbool('bool_no', True) is True
    assert c.getbool('bool_no', False) is False
    assert c.getbool('bool_no', '...') is None
    assert c.getbool('bool_true', False) is True
    assert c.getbool('bool_false', True) is False


def test_get_int():
    d = {'int_1': 1, 'int_str_1': '1', 'int_none': '...'}
    c = Settings(d)
    assert c.getint('int_1') == 1
    assert c.getint('int_str_1') == 1
    assert c.getint('int_none') is None
    assert c.getint('int_no') is None
    assert c.getint('int_1', 0) == 1
    assert c.getint('int_none', 0) is None
    assert c.getint('int_no', 0) == 0


def test_get_float():
    d = {'float_1.1': 1.1, 'float_str_1.1': '1.1', 'float_none': '...'}
    c = Settings(d)
    assert c.getfloat('float_1.1') == 1.1
    assert c.getfloat('float_str_1.1') == 1.1
    assert c.getfloat('float_none') is None
    assert c.getfloat('float_no') is None
    assert c.getfloat('float_1.1', 0) == 1.1
    assert c.getfloat('float_none', 0) is None
    assert c.getfloat('float_no', 0) == 0


def test_get_list():
    d = {'list': [1, 2], 'tuple': (1, 2), 'single': 1, 'list_str': '1,2'}
    c = Settings(d)
    assert c.getlist('list') == [1, 2]
    assert c.getlist('tuple') == [1, 2]
    assert c.getlist('single') == [1]
    assert c.getlist('list_str') == ['1', '2']
    assert c.getlist('list', [1]) == [1, 2]
    assert c.getlist('no_such_list') is None
    assert c.getlist('no_such_list', [1]) == [1]


def test_set():
    c = Settings()
    c.set('key', 'value')
    assert len(c) == 1 and c['key'] == 'value'

    c.set('key', 'value2')
    assert c['key'] == 'value2'

    c.set('key2', 'value')
    assert len(c) == 2 and c['key2'] == 'value'


def test_set_item():
    c = Settings()
    c['key'] = 'value'
    assert len(c) == 1 and c['key'] == 'value'

    c['key'] = 'value2'
    assert c['key'] == 'value2'

    c['key2'] = 'value'
    assert len(c) == 2 and c['key2'] == 'value'


def test_update():
    c = Settings()
    c.update({'key': 'value'})
    assert len(c) == 1 and c['key'] == 'value'

    c.update({'key': 'value2', 'key2': 'value'})
    assert len(c) == 2 and c['key'] == 'value2' and c['key2'] == 'value'


def test_update_by_base_c():
    c1 = Settings({'k1': 'c1_k1'})
    c2 = Settings({'k1': 'c2_k1', 'k2': 'c2_k2'}, k3='c2_k3')
    c1.update(c2)
    assert len(c1) == 3
    assert c1['k1'] == 'c2_k1'
    assert c1['k2'] == 'c2_k2'
    assert c1['k3'] == 'c2_k3'


def test_copy():
    c1 = Settings({'dict': {'k': 'v'}})
    c2 = c1.copy()
    c1['dict']['k'] = 'vv'
    assert c2['dict']['k'] == 'v'


def test_delete():
    c = Settings({'k1': 'v1', 'k2': 'v2', 'k3': 'v3'})
    c.delete('k2')
    assert len(c) == 2 and 'k2' not in c
    c.delete('k1')
    assert len(c) == 1 and 'k1' not in c
    del c['k3']
    assert len(c) == 0 and 'k3' not in c


def test_setdefault():
    c = Settings()
    c.setdefault('k1', 'v1')
    assert c['k1'] == 'v1'
    c.setdefault('k1', 'v2')
    assert c['k1'] == 'v1'



from .config import ServiceDiscoveryConfigurer
from .discovery_client import DiscoveryClient
from .load_balancer_client import LoadBalancerClient
from .load_balancer_client import RestClient
from .service_instance import ServiceInstance



from abc import ABCMeta, abstractmethod

from guniflask.security.user_details import UserDetails


class UserDetailsService(metaclass=ABCMeta):

    @abstractmethod
    def load_user_by_username(self, username: str) -> UserDetails:
        pass  # pragma: no cover



from guniflask.config import settings
from guniflask.context import configuration
from guniflask.security_config import WebSecurityConfigurer, HttpSecurity

from .jwt_config import JwtConfigurer


@configuration
class SecurityConfiguration(WebSecurityConfigurer):

    def configure_http(self, http: HttpSecurity):
        """Configure http security here"""
        jwt = settings.get_by_prefix('guniflask.jwt')
        if jwt:
            http.apply(JwtConfigurer(jwt))



import inspect
from abc import ABCMeta, abstractmethod
from functools import partial
from typing import List, get_type_hints

from guniflask.beans.constructor_resolver import ConstructorResolver
from guniflask.beans.definition import BeanDefinition
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.beans.errors import BeanTypeNotDeclaredError, BeanTypeNotAllowedError, BeanCreationError, \
    BeanNotOfRequiredTypeError, NoUniqueBeanDefinitionError
from guniflask.beans.errors import NoSuchBeanDefinitionError, BeanDefinitionStoreError
from guniflask.beans.factory import BeanFactoryAware, BeanNameAware, ConfigurableBeanFactory
from guniflask.beans.lifecycle import InitializingBean, SmartInitializingSingleton, DisposableBean
from guniflask.beans.post_processor import BeanPostProcessor


class AbstractBeanFactory(ConfigurableBeanFactory, metaclass=ABCMeta):
    def __init__(self):
        super().__init__()
        self._bean_post_processors = []

    def add_bean_post_processor(self, bean_post_processor: BeanPostProcessor):
        try:
            self._bean_post_processors.remove(bean_post_processor)
        except ValueError:
            pass
        self._bean_post_processors.append(bean_post_processor)

    @property
    def bean_post_processors(self) -> List[BeanPostProcessor]:
        return self._bean_post_processors

    def get_bean(self, bean_name, required_type: type = None):
        bean = None

        share_instance = self.get_singleton(bean_name)
        if share_instance is not None:
            bean = share_instance
        else:
            bean_definition = self.get_bean_definition(bean_name)
            if bean_definition.is_singleton():
                bean = self.get_singleton_from_factory(
                    bean_name,
                    partial(self.create_bean, bean_name, bean_definition),
                )
            elif bean_definition.is_prototype():
                # FIXME: support prototype
                raise RuntimeError('Do not support prototype now')

        # Check if required type matches the type of the actual bean instance.
        if bean is not None and required_type is not None:
            if not issubclass(type(bean), required_type):
                raise BeanNotOfRequiredTypeError(bean, required_type, type(bean))
        return bean

    def get_bean_of_type(self, required_type: type):
        candidates = self.get_beans_of_type(required_type)
        if len(candidates) == 1:
            return list(candidates.values())[0]
        if len(candidates) > 1:
            raise NoUniqueBeanDefinitionError(required_type)

    def get_beans_of_type(self, required_type: type):
        names = self.get_bean_names_for_type(required_type)
        result = {}
        for name in names:
            bean = self.get_bean(name, required_type=required_type)
            result[name] = bean
        return result

    def is_type_match(self, bean_name: str, type_to_match: type) -> bool:
        bean = self.get_singleton(bean_name)
        if bean is not None:
            return isinstance(bean, type_to_match)
        if not self.contains_bean_definition(bean_name):
            return False
        bean_type = self._resolve_bean_type(bean_name, self.get_bean_definition(bean_name))
        if bean_type is None:
            return False
        return issubclass(bean_type, type_to_match)

    def _resolve_bean_type(self, bean_name: str, bean_definition: BeanDefinition) -> type:
        source = bean_definition.source
        if inspect.isclass(source):
            return source
        if inspect.isfunction(source) or inspect.ismethod(source):
            hints = get_type_hints(source)
            if 'return' not in hints:
                raise BeanTypeNotDeclaredError(bean_name)
            bean_type = hints['return']
            if not inspect.isclass(bean_type):
                raise BeanTypeNotAllowedError(bean_name, bean_type)
            return bean_type

    @abstractmethod
    def create_bean(self, bean_name: str, bean_definition: BeanDefinition):
        pass  # pragma: no cover

    @abstractmethod
    def contains_bean_definition(self, bean_name: str) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def get_bean_definition(self, bean_name: str) -> BeanDefinition:
        pass  # pragma: no cover


class DefaultBeanFactory(AbstractBeanFactory, BeanDefinitionRegistry):
    def __init__(self):
        AbstractBeanFactory.__init__(self)
        self._constructor_resolver = ConstructorResolver(self)
        self._bean_definition_map = {}
        self._allow_bean_definition_overriding = True

    def get_bean_names_for_type(self, required_type: type) -> List[str]:
        names = []
        for bean_name in self._bean_definition_map:
            if self.is_type_match(bean_name, required_type):
                names.append(bean_name)
        return names

    def pre_instantiate_singletons(self):
        bean_names = self.get_bean_definition_names()
        for bean_name in bean_names:
            bean_definition = self.get_bean_definition(bean_name)
            if bean_definition.is_singleton():
                self.get_bean(bean_name)

        for bean_name in bean_names:
            singleton = self.get_singleton(bean_name)
            if isinstance(singleton, SmartInitializingSingleton):
                singleton.after_singletons_instantiated()

    def create_bean(self, bean_name: str, bean_definition: BeanDefinition):
        bean = self._resolve_before_instantiation(bean_name, bean_definition)
        if bean is not None:
            return bean
        bean = self._do_create_bean(bean_name, bean_definition)
        return bean

    def _do_create_bean(self, bean_name: str, bean_definition: BeanDefinition):
        bean = self._create_bean_instance(bean_name, bean_definition)
        bean = self._initialize_bean(bean, bean_name)
        self._register_disposable_bean_if_necessary(bean_name, bean, bean_definition)
        return bean

    def _create_bean_instance(self, bean_name: str, bean_definition: BeanDefinition):
        """
        1. Find the bean with the same name as arg if the required bean type is missing.
        2. Find beans which matches the required bean type.
        3. If there is only one matched bean, set it to arg.
        4. If there are more than one matched beans, choose the bean which has the same name with arg.
           Raise error if no such bean.
        5. Set arg to its declared default value.
        6. Raise error if there is any unassigned arg.

        Note: self and cls are special.
        """
        source = bean_definition.source
        factory_bean = None
        if bean_definition.factory_bean_name is not None:
            factory_bean = self.get_bean(bean_definition.factory_bean_name)
        if inspect.isclass(source):
            func = source
        elif inspect.isfunction(source) or inspect.ismethod(source):
            if factory_bean is None:
                func = source
            else:
                func = getattr(factory_bean, source.__name__)
        else:
            raise BeanCreationError(bean_name)
        try:
            bean = self._constructor_resolver.instantiate(func)
        except Exception as e:
            raise BeanCreationError(bean_name, message=f'Cannot create bean named "{bean_name}"\n{e}')
        return bean

    def _resolve_bean_type(self, bean_name: str, bean_definition: BeanDefinition) -> type:
        source = bean_definition.source
        if inspect.isclass(source):
            return source
        if inspect.isfunction(source) or inspect.ismethod(source):
            hints = get_type_hints(source)
            if 'return' not in hints:
                raise BeanTypeNotDeclaredError(bean_name)
            bean_type = hints['return']
            if not inspect.isclass(bean_type):
                raise BeanTypeNotAllowedError(bean_name, bean_type)
            return bean_type

    def _resolve_before_instantiation(self, bean_name: str, bean_definition: BeanDefinition):
        bean = None
        bean_type = self._resolve_bean_type(bean_name, bean_definition)
        if bean_type is not None:
            bean = self._apply_bean_post_processors_before_instantiation(bean_type, bean_name)
            if bean is not None:
                bean = self._apply_bean_post_processors_after_initialization(bean, bean_name)
        return bean

    def _initialize_bean(self, bean, bean_name: str):
        self._invoke_aware_methods(bean, bean_name)
        wrapped_bean = bean
        wrapped_bean = self._apply_bean_post_processors_before_initialization(wrapped_bean, bean_name)
        self._invoke_init_methods(wrapped_bean)
        wrapped_bean = self._apply_bean_post_processors_after_initialization(wrapped_bean, bean_name)
        return wrapped_bean

    def _invoke_aware_methods(self, bean, bean_name: str):
        if isinstance(bean, BeanNameAware):
            bean.set_bean_name(bean_name)
        if isinstance(bean, BeanFactoryAware):
            bean.set_bean_factory(self)

    def _invoke_init_methods(self, bean):
        if isinstance(bean, InitializingBean):
            bean.after_properties_set()

    def _apply_bean_post_processors_before_instantiation(self, bean_type: type, bean_name: str):
        for post_processor in self.bean_post_processors:
            result = post_processor.post_process_before_instantiation(bean_type, bean_name)
            if result is not None:
                return result

    def _apply_bean_post_processors_before_initialization(self, bean, bean_name: str):
        for post_processor in self.bean_post_processors:
            result = post_processor.post_process_before_initialization(bean, bean_name)
            if result is not None:
                bean = result
        return bean

    def _apply_bean_post_processors_after_initialization(self, bean, bean_name: str):
        for post_processor in self.bean_post_processors:
            result = post_processor.post_process_after_initialization(bean, bean_name)
            if result is not None:
                bean = result
        return bean

    def _register_disposable_bean_if_necessary(self, bean_name: str, bean, bean_definition: BeanDefinition):
        if bean_definition.is_singleton() and isinstance(bean, DisposableBean):
            self.register_disposable_bean(bean_name, bean)

    def set_allow_bean_definition_overriding(self, allow_bean_definition_overriding: bool):
        self._allow_bean_definition_overriding = allow_bean_definition_overriding

    @property
    def is_allow_bean_definition_overriding(self) -> bool:
        return self._allow_bean_definition_overriding

    def register_bean_definition(self, bean_name: str, bean_definition: BeanDefinition):
        old_bean_definition = self._bean_definition_map.get(bean_name)
        if old_bean_definition is not None:
            if not self.is_allow_bean_definition_overriding:
                raise BeanDefinitionStoreError(f'A bean named "{bean_name}" is already bound')
        self._bean_definition_map[bean_name] = bean_definition

    def get_bean_definition(self, bean_name: str) -> BeanDefinition:
        bean_definition = self._bean_definition_map.get(bean_name)
        if bean_definition is None:
            raise NoSuchBeanDefinitionError(bean_name)
        return bean_definition

    def get_bean_definition_names(self) -> List[str]:
        return list(self._bean_definition_map.keys())

    def contains_bean_definition(self, bean_name: str) -> bool:
        return bean_name in self._bean_definition_map

    def remove_bean_definition(self, bean_name: str):
        if bean_name not in self._bean_definition_map:
            raise NoSuchBeanDefinitionError(bean_name)
        self._bean_definition_map.pop(bean_name)



from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.errors import BadCredentialsError
from guniflask.security.password_encoder import PasswordEncoder, PlainPasswordEncoder
from guniflask.security.user_details import UserDetails
from guniflask.security.user_details_authentication_provider import UserDetailsAuthenticationProvider
from guniflask.security.user_details_service import UserDetailsService


class DaoAuthenticationProvider(UserDetailsAuthenticationProvider):
    def __init__(self, user_details_service: UserDetailsService = None, password_encoder: PasswordEncoder = None):
        super().__init__()
        self.user_details_service = user_details_service
        if password_encoder is None:
            self.password_encoder = PlainPasswordEncoder()
        else:
            self.password_encoder = password_encoder

    def retrieve_user(self, username: str, authentication: UserAuthentication) -> UserDetails:
        user = self.user_details_service.load_user_by_username(username)
        return user

    def check_authentication(self, user_details: UserDetails, authentication: UserAuthentication):
        if authentication.credentials is None:
            raise BadCredentialsError('No credentials provided')
        raw_password = str(authentication.credentials).encode('utf-8')
        if not self.password_encoder.matches(raw_password, user_details.password):
            raise BadCredentialsError('Bad credentials')



from abc import ABCMeta, abstractmethod

from guniflask.security.authentication import Authentication
from guniflask.security.user_details import UserDetails


class AuthenticationUserDetailsService(metaclass=ABCMeta):
    @abstractmethod
    def load_user_details(self, token: Authentication) -> UserDetails:
        pass  # pragma: no cover



from .annotation import async_run
from .annotation import scheduled
from .async_config import AsyncConfigurer
from .async_executor import AsyncExecutor
from .async_executor import DefaultAsyncExecutor
from .scheduling_config import SchedulingConfigurer
from .task_scheduler import DefaultTaskScheduler
from .task_scheduler import TaskScheduler



from abc import ABCMeta, abstractmethod

from guniflask.annotation import AnnotationMetadata
from guniflask.beans.definition_registry import BeanDefinitionRegistry
from guniflask.beans.factory import BeanFactory


class ConditionContext:
    def __init__(self, registry: BeanDefinitionRegistry):
        self.registry = registry
        if isinstance(registry, BeanFactory):
            self.bean_factory = registry
        else:
            self.bean_factory = None


class Condition(metaclass=ABCMeta):
    @abstractmethod
    def matches(self, context: ConditionContext, metadata: AnnotationMetadata) -> bool:
        pass  # pragma: no cover
